package cn.com.ty.lift.batch.applet.util;

import cn.com.ty.lift.batch.applet.dao.model.BaseDataModel;
import cn.com.ty.lift.batch.applet.dao.model.request.CommonRequest;
import com.google.common.collect.Maps;
import org.springframework.data.redis.core.RedisTemplate;

import java.math.BigDecimal;
import java.time.Duration;
import java.time.LocalDateTime;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

/**
 * @author huangyuan
 * @date 2020/2/4
 * @description
 */
public class DataStatisticsUtil {

    /**
     * @param
     * @return
     * @description 将数据通过公司id分组并放入到redis中
     * @date 2020/2/5 3:22 下午
     */
    public static void putStatisticsDataToRedis(List<? extends BaseDataModel> dataList, RedisTemplate redisTemplate,
                                                String dataKey) {
        //通过公司id给数组分组
        Map<Long, ? extends List<? extends BaseDataModel>> mtCompanyIdToData = dataList.stream()
                .collect(Collectors.groupingBy(BaseDataModel::getMtCompanyId));
        redisTemplate.opsForValue().set(dataKey, mtCompanyIdToData);
    }

    /**
     * @param
     * @return
     * @description 获取当前公司的数据
     * @date 2020/2/5 3:26 下午
     */
    public static List<? extends BaseDataModel> getStatisticsDataFromRedis(RedisTemplate redisTemplate, String dataKey,
                                                                           Long mtCompanyId) {
        Map<Long, ? extends List<? extends BaseDataModel>> mtCompanyIdToData =
                (Map<Long, ? extends List<? extends BaseDataModel>>) redisTemplate.opsForValue().get(dataKey);
        return mtCompanyIdToData.get(mtCompanyId);
    }

    /**
     * @param
     * @return
     * @description 获取从1月到当前月份的 月数值->条数
     * @date 2020/2/4 3:09 下午
     */
    public static Map<String, Long> initDateMap() {
        Map<String, Long> monthToCount = new HashMap<>();
        for (int i = 1; i <= LocalDateTime.now().getMonthValue(); i++) {
            //小于10的月份前面加0
            if (i < 10) {
                initDateMap().put("0" + i, 0L);
            } else {
                initDateMap().put(i + "", 0L);
            }
        }
        return monthToCount;
    }

    /**
     * @param
     * @return
     * @description 获取当前年份的最开始的时间
     * @date 2020/2/4 3:37 下午
     */
    public static LocalDateTime getYearFirstTime() {
        return LocalDateTime.of(LocalDateTime.now().getYear(), 1, 1, 0, 0, 0);
    }

    /**
     * @param localDateTime 时间数据
     * @return
     * @description 获取时间的月份值(字符串)
     * @date 2020/2/4 4:31 下午
     */
    public static String getMonthStr(LocalDateTime localDateTime) {
        //为空返回0
        if (localDateTime == null) {
            return "0";
        }
        int month = localDateTime.getMonthValue();
        if (month < 10) {
            return "0" + month;
        }
        return month + "";
    }

    /**
     * @param localDateTime 时间数据
     * @return
     * @description 获取时间的月份值(数字)
     * @date 2020/2/7 3:06 下午
     */
    public static int getMonth(LocalDateTime localDateTime) {
        //为空返回0
        if (localDateTime == null) {
            return 0;
        }
        return localDateTime.getMonthValue();
    }

    /**
     * @param afterTime  较大时间
     * @param beforeTime 较小时间
     * @return Long 差值
     * @description 请两个时间的分钟差值
     * @date 2020/2/5 10:02 上午
     */
    public static Long getDiffMinute(LocalDateTime afterTime, LocalDateTime beforeTime) {
        if (beforeTime != null && afterTime != null) {
            Duration duration = Duration.between(afterTime, beforeTime);
            return duration.toMinutes();
        }
        return 0L;
    }

    /**
     * @param
     * @return
     * @description 给请求类设置年初到现在的时间
     * @date 2020/2/7 3:51 下午
     */
    public static void setBeginYearToNow(CommonRequest commonRequest) {
        //年初时间
        commonRequest.setStartTimeDate(getYearFirstTime());
        //当前时间
        commonRequest.setEndTimeDate(LocalDateTime.now());
    }

    /**
     * @param bigDecimalValue
     * @return
     * @description 获取 BigDecimal的值
     * @date 2020/2/7 8:11 下午
     */
    public static double getDoubleValue(BigDecimal bigDecimalValue) {
        return bigDecimalValue != null ? bigDecimalValue.doubleValue() : 0d;
    }

    /**
     * 根据map的key排序
     *
     * @param map 待排序的map
     * @param isDesc 是否降序，true：降序，false：升序
     * @return 排序好的map
     */
    public static <K extends Comparable<? super K>, V> Map<K, V> sortByKey(Map<K, V> map, boolean isDesc) {
        Map<K, V> result = Maps.newLinkedHashMap();
        if (isDesc) {
            map.entrySet().stream().sorted(Map.Entry.<K, V>comparingByKey().reversed())
                    .forEachOrdered(e -> result.put(e.getKey(), e.getValue()));
        } else {
            map.entrySet().stream().sorted(Map.Entry.<K, V>comparingByKey())
                    .forEachOrdered(e -> result.put(e.getKey(), e.getValue()));
        }
        return result;
    }

    /**
     * 根据map的value排序
     *
     * @param map 待排序的map
     * @param isDesc 是否降序，true：降序，false：升序
     * @return 排序好的map
     */
    public static <K, V extends Comparable<? super V>> Map<K, V> sortByValue(Map<K, V> map, boolean isDesc) {
        Map<K, V> result = Maps.newLinkedHashMap();
        if (isDesc) {
            map.entrySet().stream().sorted(Map.Entry.<K, V>comparingByValue().reversed())
                    .forEachOrdered(e -> result.put(e.getKey(), e.getValue()));
        } else {
            map.entrySet().stream().sorted(Map.Entry.<K, V>comparingByValue())
                    .forEachOrdered(e -> result.put(e.getKey(), e.getValue()));
        }
        return result;
    }

}
