package cn.com.ty.lift.quan.config;

import cn.com.ty.lift.common.utils.Judge;
import cn.com.ty.lift.common.utils.JudgeException;
import cn.com.ty.lift.common.utils.Judger;
import cn.com.ty.lift.common.utils.StrPool;
import cn.com.xwy.boot.web.dto.RestResponse;
import cn.hutool.core.util.ArrayUtil;
import cn.hutool.core.util.StrUtil;
import com.fasterxml.jackson.databind.ObjectMapper;
import lombok.extern.slf4j.Slf4j;
import org.aspectj.lang.JoinPoint;
import org.aspectj.lang.ProceedingJoinPoint;
import org.aspectj.lang.Signature;
import org.aspectj.lang.annotation.*;
import org.aspectj.lang.reflect.MethodSignature;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.web.bind.MethodArgumentNotValidException;
import org.springframework.web.context.request.RequestContextHolder;
import org.springframework.web.context.request.ServletRequestAttributes;

import javax.servlet.http.HttpServletRequest;
import java.lang.reflect.Method;
import java.time.Duration;
import java.time.Instant;
import java.util.Arrays;
import java.util.stream.Collectors;

/**
 * @author bieao
 * @date 2019/12/3 10:30 AM
 * @description 日志切面
 */
@Slf4j
@Aspect
@Component
public class ControllerAspect {

    @Autowired
    private ObjectMapper objectMapper;
    private Instant begin = Instant.now();
    private Instant end   = Instant.now();
    private String  url;
    private String  httpMethod;
    private String  ip;
    private String  classMethod;
    private String  args;
    private String  resp;
    private long    timing;

    @Pointcut("execution(* cn.com.ty.lift.quan.*.controller..*(..))")
    public void pointcut() {
    }

    @Before("pointcut()")
    public void doBefore(JoinPoint joinPoint) throws Throwable {
        // Receives the request and get request content
        begin = Instant.now();
        ServletRequestAttributes attributes = (ServletRequestAttributes) RequestContextHolder.getRequestAttributes();
        HttpServletRequest request = attributes.getRequest();
        Signature signature = joinPoint.getSignature();

        url = StrPool.LOG_URL + request.getRequestURL().toString();
        httpMethod = StrPool.LOG_HTTP_METHOD + request.getMethod();
        ip = StrPool.LOG_IP + request.getRemoteAddr();
        classMethod = StrPool.LOG_CLASS_METHOD + signature.getDeclaringTypeName() + StrPool.DOT + signature.getName();
        args = StrPool.LOG_ARGS;
        String arg = Arrays.toString(joinPoint.getArgs()).replace("=null","");
        if(StrUtil.isNotEmpty(arg)){
            String[] argArr = arg.split(",");
            if(ArrayUtil.isNotEmpty(argArr)){
                String head = argArr[0];
                String tail = argArr[argArr.length - 1];
                if(!head.contains("=")){
                    args += head + ",";
                }
                args += Arrays.stream(argArr).filter(g -> g.contains("=")).collect(Collectors.joining(","));
                if(!tail.contains("=")){
                    args += ","+ tail;
                }
            }
        }

        judge(joinPoint);
    }

    @AfterReturning(returning = "response", pointcut = "pointcut()")
    public void doAfterReturning(RestResponse response) throws Throwable {
        // Processes the request and returns the content
        end = Instant.now();
        String res = objectMapper.writerWithDefaultPrettyPrinter().writeValueAsString(response);
        if (null != res && res.length() > 0) {
            if(!StrPool.LOG_PRINT_ALL){
                int length = res.length();
                int sub = length - StrPool.LOG_PRINT_MAX;
                int head = StrPool.LOG_PRINT_MAX >> 1;
                int tail = length - head;
                res = sub < 0 ? res : res.substring(0, head) + StrPool.NEWLINE + StrPool.NEWLINE
                        + "..." + StrPool.NEWLINE
                        + sub + " omit." + StrPool.NEWLINE
                        + "..." + StrPool.NEWLINE + StrPool.NEWLINE
                        + res.substring(tail, length);
            }
        }
        timing = Duration.between(begin, end).toMillis();
        boolean bigger = timing > StrPool.LOG_GOOD_TIME;
        resp = StrPool.LOG_RESPONSE + timing + " ms " + (bigger ? "(>_<)" : "(^_^)") + StrPool.NEWLINE + res;
        if (bigger) {
            warn();
        } else {
            info();
        }
    }

    @AfterThrowing(throwing = "ex", pointcut = "pointcut()")
    public void doAfterThrowing(Exception ex) throws Throwable {
        // Processes the exception
        end = Instant.now();
        String message = "";
        if (null != ex) {
            Class<? extends Exception> clazz = ex.getClass();
            if (clazz.equals(JudgeException.class) || clazz.equals(MethodArgumentNotValidException.class)) {
                message = ex.getMessage();
                if(clazz.equals(MethodArgumentNotValidException.class)){
                    MethodArgumentNotValidException notValid = (MethodArgumentNotValidException) ex;
                    message = notValid.getBindingResult().getFieldError().getDefaultMessage();
                }
                timing = Duration.between(begin, end).toMillis();
                resp = StrPool.LOG_RESPONSE + "(" + timing + " ms) " + message;
                error();
            }
        }
    }

    @Around("pointcut()")
    public Object interceptor(ProceedingJoinPoint proceedingJoinPoint) throws Throwable {
        return proceedingJoinPoint.proceed();
    }

    //检查必要参数非空,数值型参数必须 !null && > 0
    private void judge(JoinPoint joinPoint) {
        Object[] args = joinPoint.getArgs();
        MethodSignature methodSignature = (MethodSignature) joinPoint.getSignature();
        Method method = methodSignature.getMethod();
        Judger judger = method.getDeclaredAnnotation(Judger.class);
        if (null == judger || args.length == 0) {
            return;
        }
        Object object = args[0];
        String[] fields = judger.fields();
        if (null == object || fields.length == 0) {
            return;
        }
        Judge.judgement(object, fields);
    }

    private void info() {
        log.info(StrPool.LOG_LINE);
        log.info(url);
        log.info(httpMethod);
        log.info(ip);
        log.info(classMethod);
        log.info(args);
        log.info(resp);
        log.info(StrPool.LOG_LINE);
    }

    private void warn() {
        log.warn(StrPool.LOG_LINE);
        log.warn(url);
        log.warn(httpMethod);
        log.warn(ip);
        log.warn(classMethod);
        log.warn(args);
        log.warn(resp);
        log.warn(StrPool.LOG_LINE);
    }

    private void error() {
        log.error(StrPool.LOG_LINE);
        log.error(url);
        log.error(httpMethod);
        log.error(ip);
        log.error(classMethod);
        log.error(args);
        log.error(resp);
        log.error(StrPool.LOG_LINE);
    }
}
